module Util

using FileIO, CSV

export Path, File, savedata, execute_all, execute_single

struct Path{T} end

Base.print(io::IO, path::Path{T}) where T = write(io, joinpath(path...))

function Base.iterate(::Path{T}, state=1) where T
    state > length(T) && return nothing
    return string(T[state]), state + 1
end

struct File
    root::String
    path::Path
    name::String
end

function execute_all(suite, root)
    for method in methods(suite)
        path = method.sig.parameters[2]()
        execute_single(suite, path, root)
    end
end

function execute_single(suite, path, root)
    println("running `$(path)` ...")
    input = x -> File(root, path, x)
    output = x -> File(root, path, x)
    suite(path, input, output)
end

FileIO.load(file::File) = load(joinpath(file.root, file.path..., file.name))
FileIO.save(file::File, x) = save(joinpath(file.root, file.path..., file.name), x)
CSV.write(file::File, x) = CSV.write(joinpath(file.root, file.path..., file.name), x)

savedata(file::File; data...) =
    open(joinpath(file.root, file.path..., file.name); write=true) do io
        write_texdata(io, file.path; data...)
    end

isvalidname(x) = contains(string(x), r"^[^/\\]+$")

Base.show(io, ::MIME"text/latex", x) = print(io, x)
function Base.show(io::IO, ::MIME"text/latex", x::AbstractFloat)
    exponent = x == 0. ? 0 : floor(Int, log10(x))
    if abs(exponent) <= 3
        print(io, "\\ensuremath{", x, "}")
    else
        base = x / 10. ^ exponent
        print(io, "\\ensuremath{", base, "\\cdot 10^{", exponent, "}}")
    end
end
function Base.show(io::IO, ::MIME"text/latex", x::Integer)
    p = log10(x)
    pr = round(Int, p)
    if x == 10^pr
        print(io, "\\ensuremath{10^{", pr, "}}")
    else
        print(io, "\\ensuremath{", x, "}")
    end
end

function write_texdata(io, path; data...)
    texcmd = "Data"
    # data access tex-command
    write(io, """
        \\providecommand{\\$(texcmd)}[1]{
            \\csname $(texcmd)/#1\\endcsname
        }
        """)

    for (key, value) in pairs(data)
        isvalidname(key) && isvalidname(value) ||
            throw(ArgumentError("invalid key/value: $(key => value)"))

        # actual data encoded as tex-commands
        write(io, "\\expandafter\\def")
        write(io, "\\csname $(texcmd)/$(join(path, "/"))/$(key)\\endcsname{")
        show(io, MIME("text/latex"), value)
        write(io, "}\n")
    end
end

end
